/*
 * Ripped off from Sun's RotateFilter demo code
 */

/*
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */
import java.awt.image.ColorModel;
import java.awt.image.ImageFilter;
import java.util.Hashtable;
import java.awt.Rectangle;

public abstract class TransformFilter extends ImageFilter {

    private static ColorModel defaultRGB = ColorModel.getRGBdefault();
    protected int width, height, raster[];

    /**
     * The subclass must provide a function to compute the new value
     * of a pixel.
     */
    private abstract int pixelValue(int x, int y);
    
    public void setDimensions(int w, int h)
    {
	width = w;
	height = h;
	raster = new int[width*height];
	super.setDimensions(width, height);
    }
    
    public void setColorModel(ColorModel model)
    {
        consumer.setColorModel(defaultRGB);
    }

    public void setHints(int hintflags)
    {
        consumer.setHints(TOPDOWNLEFTRIGHT
                          | COMPLETESCANLINES
                          | SINGLEPASS
                          | (hintflags & SINGLEFRAME));
    }

    public void setPixels(int x, int y, int w, int h, ColorModel model,
                          byte pixels[], int off, int scansize)
    {
        int srcoff = off;
        int dstoff = y * width + x;
        for (int yc = 0; yc < h; yc++) {
            for (int xc = 0; xc < w; xc++) {
                raster[dstoff++] = model.getRGB(pixels[srcoff++] & 0xff);
            }
            srcoff += (scansize - w);
            dstoff += (width - w);
        }
    }

    public void setPixels(int x, int y, int w, int h, ColorModel model,
                          int pixels[], int off, int scansize)
    {
        int srcoff = off;
        int dstoff = y * width + x;
        if (model == defaultRGB) {
            for (int yc = 0; yc < h; yc++) {
                System.arraycopy(pixels, srcoff, raster, dstoff, w);
                srcoff += scansize;
                dstoff += width;
            }
        } else {
            for (int yc = 0; yc < h; yc++) {
                for (int xc = 0; xc < w; xc++) {
                    raster[dstoff++] = model.getRGB(pixels[srcoff++]);
                }
                srcoff += (scansize - w);
                dstoff += (width - w);
            }
        }
    }

    public void imageComplete(int status)
    {
        if (status == IMAGEERROR || status == IMAGEABORTED)
	{
            consumer.imageComplete(status);
            return;
        }
        int pixels[] = new int[width];
        for (int dy = 0; dy < height; dy++)
	{
            for (int dx = 0; dx < width; dx++)
	    {
		pixels[dx] = this.pixelValue(dx,dy);
            }
            consumer.setPixels(0, dy, width, 1, defaultRGB, pixels, 0, width);
        }
        consumer.imageComplete(status);
    }
}
